# -*- coding: utf-8 -*-
from __future__ import annotations

from contextlib import suppress
from typing import Any

from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QAction, QMenu, QMenuBar, QStyle, QWidget

__all__ = ['MenuBar']


class MenuBar(QMenuBar):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.menu_file: QMenu = self.addMenu(self.tr('&File'))
        self.action_load: QAction = self.menu_file.addAction(
            self._icon('document-open', 'mdi6.folder-open', standard_pixmap=QStyle.StandardPixmap.SP_DialogOpenButton),
            self.tr('&Load Catalog...'))
        self.action_reload: QAction = self.menu_file.addAction(
            self._icon('document-revert', 'mdi6.reload', standard_pixmap=QStyle.StandardPixmap.SP_BrowserReload),
            self.tr('&Reload Catalogs'))
        self.menu_file.addSeparator()
        self.action_download_catalog: QAction = self.menu_file.addAction(
            self._icon('network-receive', 'mdi6.download'),
            self.tr('&Download Catalog...'))
        self.menu_file.addSeparator()
        self.action_preferences: QAction = self.menu_file.addAction(
            self._icon('preferences-other', 'mdi6.application-settings'),
            self.tr('&Preferences...'))
        self.menu_file.addSeparator()
        self.action_quit: QAction = self.menu_file.addAction(
            self._icon('application-exit', 'mdi6.exit-run', standard_pixmap=QStyle.StandardPixmap.SP_DialogCloseButton),
            self.tr('&Quit'))

        self.menu_edit: QMenu = self.addMenu(self.tr('&Edit'))
        self.action_clear: QAction = self.menu_edit.addAction(
            self._icon('edit-clear', 'mdi6.broom', standard_pixmap=QStyle.StandardPixmap.SP_DialogResetButton),
            self.tr('&Clear Results'))
        self.menu_edit.addSeparator()
        self.menu_copy_only: QMenu = self.menu_edit.addMenu(self.tr('Copy &Only'))
        self.action_copy_current: QAction = self.menu_copy_only.addAction(self.tr('Active &Cell'))
        self.menu_copy_only.addSeparator()
        self.action_copy_name: QAction = self.menu_copy_only.addAction(self.tr('&Substance Name'))
        self.action_copy_frequency: QAction = self.menu_copy_only.addAction(self.tr('&Frequency'))
        self.action_copy_intensity: QAction = self.menu_copy_only.addAction(self.tr('&Intensity'))
        self.action_copy_lower_state_energy: QAction = self.menu_copy_only.addAction(
            self.menu_copy_only.tr('&Lower State Energy'))
        self.action_copy: QAction = self.menu_edit.addAction(
            self._icon('edit-copy', 'mdi6.content-copy'),
            self.tr('Co&py Selection'))
        self.menu_edit.addSeparator()
        self.action_select_all: QAction = self.menu_edit.addAction(
            self._icon('edit-select-all', 'mdi6.select-all'),
            self.tr('&Select All'))
        self.menu_edit.addSeparator()
        self.action_substance_info: QAction = self.menu_edit.addAction(
            self._icon('dialog-information', 'mdi6.flask-empty-outline', 'mdi6.information-variant',
                       options=[{}, {'scale_factor': 0.5}]),
            self.tr('Substance &Info'))

        self.menu_columns: QMenu = self.addMenu(self.tr('&Columns'))
        self.action_show_substance: QAction = self.menu_columns.addAction(self.tr('&Substance'))
        self.action_show_frequency: QAction = self.menu_columns.addAction(self.tr('&Frequency'))
        self.action_show_intensity: QAction = self.menu_columns.addAction(self.tr('&Intensity'))
        self.action_show_lower_state_energy: QAction = self.menu_columns.addAction(self.tr('&Lower State Energy'))

        self.menu_help: QMenu = self.addMenu(self.tr('&Help'))
        self.action_check_updates: QAction = self.menu_help.addAction(
            self._icon('application-update', 'mdi6.update'),
            self.tr('Check for Updates...'))
        self.menu_help.addSeparator()
        self.action_about_catalogs: QAction = self.menu_help.addAction(
            self._icon('document-properties', 'mdi6.information'),
            self.tr('About Catalogs...'))
        self.action_about: QAction = self.menu_help.addAction(
            self._icon('help-about', 'mdi6.help', standard_pixmap=QStyle.StandardPixmap.SP_FileDialogInfoView),
            self.tr('&About...'))
        self.action_about_qt: QAction = self.menu_help.addAction(
            self.style().standardIcon(QStyle.StandardPixmap.SP_TitleBarMenuButton),
            self.tr('About &Qt...'))

        self.action_preferences.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.action_quit.setMenuRole(QAction.MenuRole.QuitRole)
        self.action_about.setMenuRole(QAction.MenuRole.AboutRole)
        self.action_about_qt.setMenuRole(QAction.MenuRole.AboutQtRole)

        self.action_load.setShortcut('Ctrl+L')
        self.action_quit.setShortcut('Ctrl+Q')
        self.action_about.setShortcut('F1')
        self.action_preferences.setShortcut('Ctrl+,')
        self.action_copy.setShortcut('Ctrl+C')
        self.action_select_all.setShortcut('Ctrl+A')
        self.action_reload.setShortcut('Ctrl+R')
        self.action_copy_current.setShortcut('Ctrl+Shift+C')
        self.action_copy_name.setShortcut('Ctrl+Shift+C, N')
        self.action_copy_frequency.setShortcut('Ctrl+Shift+C, F')
        self.action_copy_intensity.setShortcut('Ctrl+Shift+C, I')
        self.action_copy_lower_state_energy.setShortcut('Ctrl+Shift+C, E')
        self.action_substance_info.setShortcut('Ctrl+I')
        self.action_show_substance.setCheckable(True)
        self.action_show_frequency.setCheckable(True)
        self.action_show_intensity.setCheckable(True)
        self.action_show_lower_state_energy.setCheckable(True)

    def _icon(self, theme_name: str, *qta_name: str,
              standard_pixmap: QStyle.StandardPixmap | None = None,
              **qta_specs: Any) -> QIcon:
        if theme_name and QIcon.hasThemeIcon(theme_name):
            return QIcon.fromTheme(theme_name)

        if qta_name:
            with suppress(ImportError, Exception):
                import qtawesome as qta

                return qta.icon(*qta_name, **qta_specs)  # might raise an `Exception` if the icon is not in the font

        if standard_pixmap is not None:
            return self.style().standardIcon(standard_pixmap)

        return QIcon()



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
import random
import time
from http.client import HTTPConnection, HTTPResponse, HTTPSConnection
from math import inf
from queue import Empty, Queue
from threading import Thread
from typing import Any, Final, Mapping, cast
from urllib.error import HTTPError
from urllib.parse import ParseResult, urlencode, urlparse

try:
    import orjson as json
except ImportError:
    import json

from .catalog_entry import CatalogEntry
from .utils import *

__all__ = ['Downloader', 'get_catalog', 'save_catalog', 'download']

logger: logging.Logger = logging.getLogger('downloader')


class Downloader(Thread):
    def __init__(self,
                 frequency_limits: tuple[float, float] = (-inf, inf),
                 state_queue: Queue[tuple[int, int]] | None = None) -> None:
        super().__init__()
        self._state_queue: Queue[tuple[int, int]] | None = state_queue
        self._frequency_limits: tuple[float, float] = frequency_limits
        self._catalog: list[dict[str, int | str | list[dict[str, float]]]] = []

        self._run: bool = False
        self._sessions: dict[tuple[str, str], HTTPConnection | HTTPSConnection] = dict()

    def __del__(self) -> None:
        self._run = False
        session: HTTPConnection | HTTPSConnection
        for session in self._sessions.values():
            session.close()

    @property
    def catalog(self) -> list[dict[str, int | str | list[dict[str, float]]]]:
        return self._catalog.copy()

    def join(self, timeout: float | None = ...) -> None:
        self._run = False
        session: HTTPConnection | HTTPSConnection
        for session in self._sessions.values():
            session.close()
        super().join(timeout=timeout)

    def run(self) -> None:
        self._run = True

        def session_for_url(scheme: str, location: str) -> HTTPConnection | HTTPSConnection:
            if (scheme, location) not in self._sessions:
                if scheme == 'http':
                    self._sessions[(scheme, location)] = HTTPConnection(location)
                elif scheme == 'https':
                    self._sessions[(scheme, location)] = HTTPSConnection(location)
                else:
                    raise ValueError(f'Unknown scheme: {scheme}')
            return self._sessions[(scheme, location)]

        def get(url: str, headers: Mapping[str, str] | None = None) -> str:
            parse_result: ParseResult = urlparse(url)
            session: HTTPConnection | HTTPSConnection = session_for_url(parse_result.scheme, parse_result.netloc)
            response: HTTPResponse
            while True:
                try:
                    session.request(method='GET', url=parse_result.path, headers=(headers or dict()))
                    response = session.getresponse()
                except ConnectionResetError:
                    time.sleep(random.random())
                else:
                    break
            if response.closed:
                return ''
            try:
                return response.read().decode()
            except AttributeError:  # `response.fp` became `None` before socket began closing
                return ''

        def post(url: str, data: dict[str, Any], headers: Mapping[str, str] | None = None) -> str:
            parse_result: ParseResult = urlparse(url)
            session: HTTPConnection | HTTPSConnection = session_for_url(parse_result.scheme, parse_result.netloc)
            response: HTTPResponse
            while True:
                try:
                    session.request(method='POST', url=parse_result.path, body=urlencode(data),
                                    headers=(headers or dict()))
                    response = session.getresponse()
                except ConnectionResetError:
                    time.sleep(random.random())
                else:
                    break
            if response.closed:
                return ''
            try:
                return response.read().decode()
            except AttributeError:  # `response.fp` became `None` before socket began closing
                return ''

        def get_species() -> list[dict[str, int | str]]:
            def purge_null_data(entry: dict[str, None | int | str]) -> dict[str, int | str]:
                key: str
                value: None | int | str
                return dict(
                    (key, value) for key, value in entry.items() if value is not None and value not in ('', 'None'))

            def trim_strings(entry: dict[str, None | int | str]) -> dict[str, None | int | str]:
                key: str
                for key in entry:
                    if isinstance(entry[key], str):
                        entry[key] = cast(str, entry[key]).strip()
                return entry

            data: dict[str, int | str | list[dict[str, None | int | str]]] \
                = json.loads(post('https://cdms.astro.uni-koeln.de/cdms/portal/json_list/species/', {'database': -1},
                                  headers={'Content-Type': 'application/x-www-form-urlencoded'}))
            if 'species' in data:
                return [purge_null_data(trim_strings(s)) for s in data['species']]
            else:
                return []

        def get_substance_catalog(species_entry: dict[str, int | str]) -> dict[str, int | str | list[dict[str, float]]]:
            if not self._run:
                return dict()  # quickly exit the function

            def entry_url(species_tag: int) -> str:
                entry_filename: str = f'c{species_tag:06}.cat'
                if entry_filename in ('c044009.cat', 'c044012.cat'):  # merged with c044004.cat — Brian J. Drouin
                    return ''
                if species_tag % 1000 > 500:
                    return 'https://cdms.astro.uni-koeln.de/classic/entries/' + entry_filename
                else:
                    return 'https://spec.jpl.nasa.gov/ftp/pub/catalog/' + entry_filename

            if SPECIES_TAG not in species_entry:
                # nothing to go on with
                return dict()
            fn: str = entry_url(species_tag=cast(int, species_entry[SPECIES_TAG]))
            if not fn:  # no need to download a file for the species tag
                return dict()
            try:
                logger.debug(f'getting {fn}')
                lines = get(fn).splitlines()
            except HTTPError as ex:
                logger.error(fn, exc_info=ex)
                return dict()
            catalog_entries = [CatalogEntry(line) for line in lines]
            if not catalog_entries:
                logger.warning('no entries in the catalog')
                return dict()
            return {
                **species_entry,
                DEGREES_OF_FREEDOM: catalog_entries[0].degrees_of_freedom,
                LINES: [_catalog_entry.to_dict()
                        for _catalog_entry in catalog_entries
                        if within(_catalog_entry.frequency, self._frequency_limits)]
            }

        species: list[dict[str, int | str]] = get_species()
        catalog: list[dict[str, int | str | list[dict[str, float]]]] = []
        species_count: Final[int] = len(species)
        skipped_count: int = 0
        catalog_entry: dict[str, int | str | list[dict[str, float]]]
        entry_index: int
        _e: dict[str, int | str]
        for entry_index, _e in enumerate(species):
            catalog_entry = get_substance_catalog(_e)
            if catalog_entry and LINES in catalog_entry and catalog_entry[LINES]:
                catalog.append(catalog_entry)
            else:
                skipped_count += 1
            if self._state_queue is not None:
                self._state_queue.put((len(catalog), species_count - entry_index - skipped_count))

        if self._state_queue is not None:
            self._state_queue.put((len(catalog), 0))

        self._catalog = catalog


def get_catalog(frequency_limits: tuple[float, float] = (-inf, inf)) \
        -> list[dict[str, int | str | list[dict[str, float]]]]:
    """
    Download the spectral lines catalog data

    :param tuple frequency_limits: the frequency range of the catalog entries to keep.
    :return: a list of the spectral lines catalog entries.
    """

    state_queue: Queue[tuple[int, int]] = Queue()
    downloader: Downloader = Downloader(frequency_limits=frequency_limits, state_queue=state_queue)
    downloader.start()
    while downloader.is_alive():
        cataloged_species: int
        not_yet_processed_species: int
        try:
            cataloged_species, not_yet_processed_species = state_queue.get(block=True, timeout=0.1)
        except Empty:
            continue
        except KeyboardInterrupt:
            downloader.join(0.1)
        else:
            logger.info(f'got {cataloged_species} entries, {not_yet_processed_species} left')

    return downloader.catalog


def save_catalog(filename: str,
                 frequency_limits: tuple[float, float] = (0, inf)) -> bool:
    """
    Download and save the spectral lines catalog data

    :param str filename: the name of the file to save the downloaded catalog to.
        If it ends with an unknown suffix, `'.json.gz'` is appended to it.
    :param tuple frequency_limits: the tuple of the maximal and the minimal frequencies of the lines being stored.
        All the lines outside the specified frequency range are omitted.
    """

    return save_catalog_to_file(filename=filename,
                                catalog=get_catalog(frequency_limits),
                                frequency_limits=frequency_limits)


def download() -> None:
    import argparse
    from datetime import datetime

    ap: argparse.ArgumentParser = argparse.ArgumentParser(
        allow_abbrev=True,
        description='Download JPL and CDMS spectroscopy catalogs for offline search.\n'
                    'Find more at https://github.com/StSav012/pycatsearch.')
    ap.add_argument('catalog', type=str, help='the catalog location to save into (required)')
    ap.add_argument('-f''min', '--min-frequency', type=float, help='the lower frequency [MHz] to take', default=-inf)
    ap.add_argument('-f''max', '--max-frequency', type=float, help='the upper frequency [MHz] to take', default=+inf)
    args: argparse.Namespace = ap.parse_intermixed_args()

    logging.basicConfig(level=logging.DEBUG)
    logger.info(f'started at {datetime.now()}')
    save_catalog(args.catalog, (args.min_frequency, args.max_frequency))
    logger.info(f'finished at {datetime.now()}')



# coding=utf-8
from __future__ import annotations

import site

from setuptools import setup

site.main()


def required_packages() -> list[str]:
    import sys
    if sys.version_info < (3, 8):
        raise SystemError('Python versions prior to 3.8 are not supported')

    import platform
    from typing import NamedTuple, Sequence

    from pkg_resources import parse_version

    class PackageRequirement(NamedTuple):
        package_name: str
        min_version: str = ''

        def __str__(self) -> str:
            if self.min_version:
                return self.package_name + ' >= ' + self.min_version
            return self.package_name

    def is_package_importable(package_requirement: PackageRequirement) -> bool:
        from importlib.metadata import version, PackageNotFoundError

        try:
            version(package_requirement.package_name)
        except PackageNotFoundError:
            return False
        else:
            if (package_requirement.min_version
                    and (parse_version(version(package_requirement.package_name))
                         < parse_version(package_requirement.min_version))):
                return False
        return True

    def required_package(package_requirement: PackageRequirement | Sequence[PackageRequirement]) -> PackageRequirement:
        """
        Install packages if missing

        :param package_requirement: a package name or a sequence of the names of alternative packages;
                             if none of the packages installed beforehand, install the first one given
        :returns bool: True if a package is importable, False when an attempt to install the package made
        """

        if not package_requirement:
            raise ValueError('No package requirements given')

        if not isinstance(package_requirement, PackageRequirement) and isinstance(package_requirement, Sequence):
            for _package_requirement in package_requirement:
                if is_package_importable(_package_requirement):
                    return _package_requirement
            return package_requirement[0]

        if isinstance(package_requirement, PackageRequirement):
            if is_package_importable(package_requirement):
                return package_requirement
            else:
                return package_requirement
        raise ValueError('Invalid requirements')

    qt_list: list[PackageRequirement]
    uname: platform.uname_result = platform.uname()
    if ((uname.system == 'Windows'
         and parse_version(uname.version) < parse_version('10.0.19044'))  # Windows 10 21H2 or later required
            or uname.machine not in ('x86_64', 'AMD64')):
        # Qt6 does not support the OSes
        qt_list = [
            PackageRequirement(package_name='PyQt5', min_version='5.15.0'),
            PackageRequirement(package_name='PySide2', min_version='5.15.0'),
        ]
    else:
        qt_list = [
            PackageRequirement(package_name='PySide6-Essentials', min_version='6.2.0'),
            PackageRequirement(package_name='PyQt6', min_version='6.2.0'),
            PackageRequirement(package_name='PyQt5', min_version='5.15.0'),
            PackageRequirement(package_name='PySide2', min_version='5.15.0'),
        ]

    requirements: list[PackageRequirement | Sequence[PackageRequirement]] = [
        PackageRequirement(package_name='qtpy', min_version='2.3.1'),
        qt_list,
    ]
    return [str(required_package(requirement)) for requirement in requirements]


setup(install_requires=required_packages())



# -*- coding: utf-8 -*-
from __future__ import annotations

from qtpy.QtWidgets import QWidget

from .selectable_label import SelectableLabel


__all__ = ['URLLabel']


class URLLabel(SelectableLabel):
    """ A label with selectable hyperlink """

    def __init__(self, url: str, text: str, parent: QWidget | None = None) -> None:
        super().__init__(f'<a href="{url}">{text}</a>', parent)
        self.setOpenExternalLinks(True)
        self.setToolTip(url)



# -*- coding: utf-8 -*-
from __future__ import annotations

import logging
from functools import partial
from typing import Any, Hashable, cast

from qtpy.QtCore import QByteArray, Qt
from qtpy.QtGui import QCloseEvent
from qtpy.QtWidgets import (QCheckBox, QComboBox, QDialog, QDialogButtonBox, QDoubleSpinBox, QFormLayout, QHBoxLayout,
                            QListWidget, QScrollArea, QSpinBox, QStackedWidget, QVBoxLayout, QWidget, QListWidgetItem)

from .settings import Settings

__all__ = ['Preferences']


class PreferencesPage(QWidget):
    """ A page of the Preferences dialog """

    def __init__(self, value: dict[str, (Settings.CallbackOnly
                                         | Settings.SpinboxAndCallback
                                         | Settings.ComboboxAndCallback
                                         | Settings.EditableComboboxAndCallback)],
                 settings: Settings, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.settings: Settings = settings
        logger: logging.Logger = logging.getLogger('preferences')

        if not (isinstance(value, dict) and value):
            raise TypeError(f'Invalid type: {type(value)}')
        layout: QFormLayout = QFormLayout(self)
        key2: str
        value2: (Settings.CallbackOnly
                 | Settings.SpinboxAndCallback
                 | Settings.ComboboxAndCallback
                 | Settings.EditableComboboxAndCallback)

        check_box: QCheckBox
        spin_box: QSpinBox | QDoubleSpinBox
        combo_box: QComboBox

        for key2, value2 in value.items():
            if isinstance(value2, Settings.CallbackOnly):
                if isinstance(getattr(self.settings, value2.callback), bool):
                    check_box = QCheckBox(self.tr(key2), self)
                    setattr(check_box, 'callback', value2.callback)
                    check_box.setChecked(getattr(self.settings, value2.callback))
                    check_box.toggled.connect(partial(self._on_event, sender=check_box))
                    layout.addWidget(check_box)
                else:
                    logger.error(f'The type of {value2.callback!r} is not supported')
            elif isinstance(value2, Settings.SpinboxAndCallback):
                if isinstance(getattr(self.settings, value2.callback), int):
                    spin_box = QSpinBox(self)
                else:
                    spin_box = QDoubleSpinBox(self)
                spin_box.setValue(getattr(self.settings, value2.callback))
                spin_box.setRange(value2.range.start, value2.range.stop)
                spin_box.setSingleStep(value2.range.step or 1)
                spin_box.setPrefix(value2.prefix_and_suffix[0])
                spin_box.setSuffix(value2.prefix_and_suffix[1])
                setattr(spin_box, 'callback', value2.callback)
                spin_box.valueChanged.connect(partial(self._on_event, sender=spin_box))
                layout.addRow(key2, spin_box)
            elif isinstance(value2, Settings.ComboboxAndCallback):
                combo_box = QComboBox(self)
                setattr(combo_box, 'callback', value2.callback)
                combobox_data: dict[Hashable, str]
                if isinstance(value2.combobox_data, dict):
                    combobox_data = value2.combobox_data
                else:
                    combobox_data = dict(enumerate(value2.combobox_data))
                for index, (data, item) in enumerate(combobox_data.items()):
                    combo_box.addItem(self.tr(item), data)
                combo_box.setEditable(False)
                combo_box.setCurrentText(combobox_data[getattr(self.settings, value2.callback)])
                combo_box.currentIndexChanged.connect(
                    partial(self._on_combo_box_current_index_changed, sender=combo_box))
                layout.addRow(self.tr(key2), combo_box)
            elif isinstance(value2, Settings.EditableComboboxAndCallback):
                combo_box = QComboBox(self)
                setattr(combo_box, 'callback', value2.callback)
                combo_box.addItems(value2.combobox_items)
                current_text: str = getattr(self.settings, value2.callback)
                if current_text in value2.combobox_items:
                    combo_box.setCurrentIndex(value2.combobox_items.index(current_text))
                else:
                    combo_box.insertItem(0, current_text)
                    combo_box.setCurrentIndex(0)
                combo_box.setEditable(True)
                combo_box.currentTextChanged.connect(partial(self._on_event, sender=combo_box))
                layout.addRow(self.tr(key2), combo_box)
            else:
                logger.error(f'{value2!r} is not supported')

    # https://forum.qt.io/post/671245
    def _on_event(self, x: bool | int | float | str, sender: QWidget) -> None:
        setattr(self.settings, getattr(sender, 'callback'), x)

    def _on_combo_box_current_index_changed(self, _: int, sender: QComboBox) -> None:
        setattr(self.settings, getattr(sender, 'callback'), sender.currentData())


class PreferencesBody(QScrollArea):
    """ The main area of the GUI preferences dialog """

    def __init__(self, settings: Settings, parent: QWidget | None = None) -> None:
        from . import icon  # import locally to avoid a circular import

        super().__init__(parent)

        self.settings: Settings = settings
        logger: logging.Logger = logging.getLogger('preferences')

        widget: QWidget = QWidget(self)
        self.setWidget(widget)
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setFrameStyle(0)

        layout: QHBoxLayout = QHBoxLayout(widget)
        content: QListWidget = QListWidget(widget)
        stack: QStackedWidget = QStackedWidget(widget)
        key: str | tuple[str, tuple[str, ...]] | tuple[str, tuple[str, ...], tuple[tuple[str, Any], ...]]
        value: dict[str, (Settings.CallbackOnly
                          | Settings.SpinboxAndCallback
                          | Settings.ComboboxAndCallback
                          | Settings.EditableComboboxAndCallback)]
        for key, value in self.settings.dialog.items():
            if not (isinstance(value, dict) and value):
                logger.error(f'Invalid value of {key!r}')
                continue
            new_item: QListWidgetItem
            if isinstance(key, str):
                new_item = QListWidgetItem(key)
            elif isinstance(key, tuple):
                if len(key) == 1:
                    new_item = QListWidgetItem(key[0])
                elif len(key) == 2:
                    new_item = QListWidgetItem(icon(*key[1]), key[0])
                elif len(key) == 3:
                    new_item = QListWidgetItem(icon(*key[1], **dict(key[2])), key[0])
                else:
                    logger.error(f'Invalid key: {key!r}')
                    continue
            else:
                logger.error(f'Invalid key type: {key!r}')
                continue
            content.addItem(new_item)
            box: PreferencesPage = PreferencesPage(value, settings, self)
            stack.addWidget(box)
        content.setMinimumWidth(content.sizeHintForColumn(0) + 2 * content.frameWidth())
        layout.addWidget(content)
        layout.addWidget(stack)

        if content.count() > 0:
            content.setCurrentRow(0)  # select the first page

        content.currentRowChanged.connect(stack.setCurrentIndex)


class Preferences(QDialog):
    """ GUI preferences dialog """

    def __init__(self, settings: Settings, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.settings: Settings = settings
        self.setModal(True)
        self.setWindowTitle(self.tr('Preferences'))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())

        layout: QVBoxLayout = QVBoxLayout(self)
        layout.addWidget(PreferencesBody(settings=settings, parent=parent))
        buttons: QDialogButtonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Close, self)
        buttons.rejected.connect(self.close)
        layout.addWidget(buttons)

        self.settings.beginGroup('PreferencesDialog')
        self.restoreGeometry(cast(QByteArray, self.settings.value('windowGeometry', QByteArray())))
        self.settings.endGroup()

    def closeEvent(self, event: QCloseEvent) -> None:
        self.settings.beginGroup('PreferencesDialog')
        self.settings.setValue('windowGeometry', self.saveGeometry())
        self.settings.endGroup()



# -*- coding: utf-8 -*-
from __future__ import annotations

import os
from typing import Any, Callable, Final, Hashable, Iterable, NamedTuple, Sequence

from qtpy.QtCore import QObject, QSettings

from ..utils import *

__all__ = ['Settings']


class Settings(QSettings):
    """ convenient internal representation of the application settings """

    class CallbackOnly(NamedTuple):
        callback: str

    class SpinboxAndCallback(NamedTuple):
        range: slice
        prefix_and_suffix: tuple[str, str]
        callback: str

    class ComboboxAndCallback(NamedTuple):
        combobox_data: Iterable[str] | dict[Hashable, str]
        callback: str

    class EditableComboboxAndCallback(NamedTuple):
        combobox_items: Sequence[str]
        callback: str

    TO_MHZ: Final[list[Callable[[float], float]]] = [lambda x: x, ghz_to_mhz, rec_cm_to_mhz, nm_to_mhz]
    FROM_MHZ: Final[list[Callable[[float], float]]] = [lambda x: x, mhz_to_ghz, mhz_to_rec_cm, mhz_to_nm]

    TO_LOG10_SQ_NM_MHZ: Final[list[Callable[[float], float]]] = [
        lambda x: x,
        sq_nm_mhz_to_log10_sq_nm_mhz,
        log10_cm_per_molecule_to_log10_sq_nm_mhz,
        cm_per_molecule_to_log10_sq_nm_mhz
    ]
    FROM_LOG10_SQ_NM_MHZ: Final[list[Callable[[float], float]]] = [
        lambda x: x,
        log10_sq_nm_mhz_to_sq_nm_mhz,
        log10_sq_nm_mhz_to_log10_cm_per_molecule,
        log10_sq_nm_mhz_to_cm_per_molecule
    ]

    TO_REC_CM: Final[list[Callable[[float], float]]] = [
        lambda x: x,
        meV_to_rec_cm,
        j_to_rec_cm
    ]
    FROM_REC_CM: Final[list[Callable[[float], float]]] = [
        lambda x: x,
        rec_cm_to_meV,
        rec_cm_to_j
    ]

    TO_K: Final[list[Callable[[float], float]]] = [lambda x: x, lambda x: x + 273.15]
    FROM_K: Final[list[Callable[[float], float]]] = [lambda x: x, lambda x: x - 273.15]

    INCHI_KEY_SEARCH_PROVIDERS: Final[list[str]] = [
        'https://pubchem.ncbi.nlm.nih.gov/#query={InChIKey}',
        'https://www.ebi.ac.uk/unichem/compoundsources?type=inchikey&compound={InChIKey}',
        'https://webbook.nist.gov/cgi/cbook.cgi?InChI={InChIKey}',
        'https://www.spectrabase.com/search?q={InChIKey}',
        'https://www.google.com/search?q={InChIKey}',
        'http://gmd.mpimp-golm.mpg.de/search.aspx?query={InChIKey}',
        'http://www.chemspider.com/InChIKey/{InChIKey}',
    ]

    def __init__(self, organization: str, application: str, parent: QObject | None = None) -> None:
        super().__init__(organization, application, parent)

        # for some reason, the dicts are not being translated when used as class variables
        self.LINE_ENDS: Final[dict[str, str]] = {
            '\n': self.tr(r'Line Feed (\n)'),
            '\r': self.tr(r'Carriage Return (\r)'),
            '\r\n': self.tr(r'CR+LF (\r\n)'),
            '\n\r': self.tr(r'LF+CR (\n\r)')
        }
        self.CSV_SEPARATORS: Final[dict[str, str]] = {
            ',': self.tr(r'comma (,)'),
            '\t': self.tr(r'tab (\t)'),
            ';': self.tr(r'semicolon (;)'),
            ' ': self.tr(r'space ( )')
        }
        self.FREQUENCY_UNITS: Final[list[str]] = [self.tr('MHz'), self.tr('GHz'), self.tr('cm⁻¹'), self.tr('nm')]
        self.INTENSITY_UNITS: Final[list[str]] = [self.tr('lg(nm² × MHz)'), self.tr('nm² × MHz'),
                                                  self.tr('lg(cm / molecule)'), self.tr('cm / molecule')]
        self.ENERGY_UNITS: Final[list[str]] = [self.tr('cm⁻¹'), self.tr('meV'), self.tr('J')]
        self.TEMPERATURE_UNITS: Final[list[str]] = [self.tr('K'), self.tr('°C')]

    @property
    def dialog(self) -> (dict[(str
                               | tuple[str, tuple[str, ...]]
                               | tuple[str, tuple[str, ...], tuple[tuple[str, Any], ...]]),
    dict[str, (Settings.CallbackOnly
               | Settings.SpinboxAndCallback
               | Settings.ComboboxAndCallback)]]):
        return {
            (self.tr('When the program starts'), ('mdi6.rocket-launch',)): {
                self.tr('Load catalogs'): Settings.CallbackOnly('load_last_catalogs'),
                self.tr('Check for update'): Settings.CallbackOnly('check_updates'),
            },
            (self.tr('Display'), ('mdi6.binoculars',)): {
                self.tr('Allow rich text in formulas'): Settings.CallbackOnly('rich_text_in_formulas'),
            },
            (self.tr('Search'), ('mdi6.table-search',)): {
                self.tr('Timeout:'): Settings.SpinboxAndCallback(slice(1, 99), ('', 'sec'), 'timeout'),
            },
            (self.tr('Units'), ('mdi6.pencil-ruler',)): {
                self.tr('Frequency:'): Settings.ComboboxAndCallback(self.FREQUENCY_UNITS, 'frequency_unit'),
                self.tr('Intensity:'): Settings.ComboboxAndCallback(self.INTENSITY_UNITS, 'intensity_unit'),
                self.tr('Energy:'): Settings.ComboboxAndCallback(self.ENERGY_UNITS, 'energy_unit'),
                self.tr('Temperature:'): Settings.ComboboxAndCallback(self.TEMPERATURE_UNITS, 'temperature_unit'),
            },
            (self.tr('Export'), ('mdi6.file-export',)): {
                self.tr('With units'): Settings.CallbackOnly('with_units'),
                self.tr('Line ending:'): Settings.ComboboxAndCallback(self.LINE_ENDS, 'line_end'),
                self.tr('CSV separator:'): Settings.ComboboxAndCallback(self.CSV_SEPARATORS, 'csv_separator'),
            },
            (
                self.tr('Info'),
                ('mdi6.flask-empty-outline', 'mdi6.information-variant'),
                (('options', ((), (('scale_factor', 0.5),))),)
            ): {
                self.tr('InChI key search URL:'): Settings.EditableComboboxAndCallback(self.INCHI_KEY_SEARCH_PROVIDERS,
                                                                                       'inchi_key_search_url_template'),
            },
        }

    @property
    def frequency_unit(self) -> int:
        self.beginGroup('frequency')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return v

    @frequency_unit.setter
    def frequency_unit(self, new_value: int | str) -> None:
        self.beginGroup('frequency')
        if isinstance(new_value, str):
            new_value = self.FREQUENCY_UNITS.index(new_value)
        self.setValue('unit', new_value)
        self.endGroup()

    @property
    def frequency_unit_str(self) -> str:
        self.beginGroup('frequency')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.FREQUENCY_UNITS[v]

    @property
    def to_mhz(self) -> Callable[[float], float]:
        self.beginGroup('frequency')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.TO_MHZ[v]

    @property
    def from_mhz(self) -> Callable[[float], float]:
        self.beginGroup('frequency')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.FROM_MHZ[v]

    @property
    def intensity_unit(self) -> int:
        self.beginGroup('intensity')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return v

    @intensity_unit.setter
    def intensity_unit(self, new_value: int | str) -> None:
        self.beginGroup('intensity')
        if isinstance(new_value, str):
            new_value = self.INTENSITY_UNITS.index(new_value)
        self.setValue('unit', new_value)
        self.endGroup()

    @property
    def intensity_unit_str(self) -> str:
        self.beginGroup('intensity')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.INTENSITY_UNITS[v]

    @property
    def to_log10_sq_nm_mhz(self) -> Callable[[float], float]:
        self.beginGroup('intensity')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.TO_LOG10_SQ_NM_MHZ[v]

    @property
    def from_log10_sq_nm_mhz(self) -> Callable[[float], float]:
        self.beginGroup('intensity')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.FROM_LOG10_SQ_NM_MHZ[v]

    @property
    def energy_unit(self) -> int:
        self.beginGroup('energy')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return v

    @energy_unit.setter
    def energy_unit(self, new_value: int | str) -> None:
        self.beginGroup('energy')
        if isinstance(new_value, str):
            new_value = self.ENERGY_UNITS.index(new_value)
        self.setValue('unit', new_value)
        self.endGroup()

    @property
    def energy_unit_str(self) -> str:
        self.beginGroup('energy')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.ENERGY_UNITS[v]

    @property
    def to_rec_cm(self) -> Callable[[float], float]:
        self.beginGroup('energy')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.TO_REC_CM[v]

    @property
    def from_rec_cm(self) -> Callable[[float], float]:
        self.beginGroup('energy')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.FROM_REC_CM[v]

    @property
    def temperature_unit(self) -> int:
        self.beginGroup('temperature')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return v

    @temperature_unit.setter
    def temperature_unit(self, new_value: int | str) -> None:
        self.beginGroup('temperature')
        if isinstance(new_value, str):
            new_value = self.TEMPERATURE_UNITS.index(new_value)
        self.setValue('unit', new_value)
        self.endGroup()

    @property
    def temperature_unit_str(self) -> str:
        self.beginGroup('temperature')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.TEMPERATURE_UNITS[v]

    @property
    def to_k(self) -> Callable[[float], float]:
        self.beginGroup('temperature')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.TO_K[v]

    @property
    def from_k(self) -> Callable[[float], float]:
        self.beginGroup('temperature')
        v: int = self.value('unit', 0, int)
        self.endGroup()
        return self.FROM_K[v]

    @property
    def load_last_catalogs(self) -> bool:
        self.beginGroup('start')
        v: bool = self.value('loadLastCatalogs', True, bool)
        self.endGroup()
        return v

    @load_last_catalogs.setter
    def load_last_catalogs(self, new_value: bool) -> None:
        self.beginGroup('start')
        self.setValue('loadLastCatalogs', new_value)
        self.endGroup()

    @property
    def check_updates(self) -> bool:
        self.beginGroup('start')
        v: bool = self.value('checkUpdates', True, bool)
        self.endGroup()
        return v

    @check_updates.setter
    def check_updates(self, new_value: bool) -> None:
        self.beginGroup('start')
        self.setValue('checkUpdates', new_value)
        self.endGroup()

    @property
    def rich_text_in_formulas(self) -> bool:
        self.beginGroup('display')
        v: bool = self.value('richTextInFormulas', True, bool)
        self.endGroup()
        return v

    @rich_text_in_formulas.setter
    def rich_text_in_formulas(self, new_value: bool) -> None:
        self.beginGroup('display')
        self.setValue('richTextInFormulas', new_value)
        self.endGroup()

    @property
    def line_end(self) -> str:
        self.beginGroup('export')
        v: int = self.value('lineEnd', list(self.LINE_ENDS.keys()).index(os.linesep), int)
        self.endGroup()
        return list(self.LINE_ENDS.keys())[v]

    @line_end.setter
    def line_end(self, new_value: str) -> None:
        self.beginGroup('export')
        self.setValue('lineEnd', list(self.LINE_ENDS.keys()).index(new_value))
        self.endGroup()

    @property
    def csv_separator(self) -> str:
        self.beginGroup('export')
        v: int = self.value('csvSeparator', list(self.CSV_SEPARATORS.keys()).index('\t'), int)
        self.endGroup()
        return list(self.CSV_SEPARATORS.keys())[v]

    @csv_separator.setter
    def csv_separator(self, new_value: str) -> None:
        self.beginGroup('export')
        self.setValue('csvSeparator', list(self.CSV_SEPARATORS.keys()).index(new_value))
        self.endGroup()

    @property
    def with_units(self) -> bool:
        self.beginGroup('export')
        v: bool = self.value('withUnits', True, bool)
        self.endGroup()
        return v

    @with_units.setter
    def with_units(self, new_value: bool) -> None:
        self.beginGroup('export')
        self.setValue('withUnits', new_value)
        self.endGroup()

    @property
    def timeout(self) -> float:
        self.beginGroup('search')
        v: float = self.value('timeout', 99.0, float)
        self.endGroup()
        return v

    @timeout.setter
    def timeout(self, new_value: float) -> None:
        self.beginGroup('search')
        self.setValue('timeout', new_value)
        self.endGroup()

    @property
    def ignored_version(self) -> str:
        self.beginGroup('update')
        v: str = self.value('ignoredVersion', '', str)
        self.endGroup()
        return v

    @ignored_version.setter
    def ignored_version(self, new_value: str) -> None:
        self.beginGroup('update')
        self.setValue('ignoredVersion', new_value)
        self.endGroup()

    @property
    def inchi_key_search_url_template(self) -> str:
        self.beginGroup('info')
        v: str = self.value('InChIKeySearchURLTemplate', 'https://pubchem.ncbi.nlm.nih.gov/#query={InChIKey}', str)
        self.endGroup()
        return v

    @inchi_key_search_url_template.setter
    def inchi_key_search_url_template(self, new_value: str) -> None:
        self.beginGroup('info')
        self.setValue('InChIKeySearchURLTemplate', new_value)
        self.endGroup()



# -*- coding: utf-8 -*-
from __future__ import annotations

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QLabel, QWidget

__all__ = ['SelectableLabel']


class SelectableLabel(QLabel):
    """ A label with selectable text """

    def __init__(self, text: str, parent: QWidget | None = None) -> None:
        super().__init__(text, parent)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)



# -*- coding: utf-8 -*-
from __future__ import annotations

import bz2
import gzip
import lzma
import math
import os.path
import time
from contextlib import contextmanager
from datetime import datetime, timezone
from os import PathLike
from pathlib import Path
from typing import Any, AnyStr, BinaryIO, Callable, Dict, List, NamedTuple, Optional, TextIO, Union, cast

try:
    import orjson as json
except ImportError:
    import json

from .utils import *

__all__ = ['Catalog', 'CatalogSourceInfo', 'LineType', 'LinesType', 'CatalogEntryType']

LineType = Dict[str, float]
LinesType = List[LineType]
CatalogEntryType = Dict[str, Union[int, str, LinesType]]


class CatalogSourceInfo(NamedTuple):
    filename: str
    build_datetime: datetime | None = None


class CatalogData:
    def __init__(self) -> None:
        self.catalog: list[CatalogEntryType] = []
        self.frequency_limits: tuple[tuple[float, float], ...] = ()

    def append(self,
               catalog: list[CatalogEntryType],
               frequency_limits: tuple[float, float]) -> None:
        def squash_same_species_tag_entries() -> None:
            i: int = 0
            while i < len(self.catalog) - 1:
                while i < len(self.catalog) - 1 and self.catalog[i][SPECIES_TAG] == self.catalog[i + 1][SPECIES_TAG]:
                    self.catalog[i][LINES] = cast(LinesType,
                                                  merge_sorted(self.catalog[i][LINES],
                                                               self.catalog[i + 1][LINES],
                                                               key=lambda line: (line[FREQUENCY],
                                                                                 line[INTENSITY],
                                                                                 line[LOWER_STATE_ENERGY])
                                                               )
                                                  )
                    del self.catalog[i + 1]
                i += 1

        def merge_frequency_tuples(*args: tuple[float, float] | list[float]) -> tuple[tuple[float, float], ...]:
            if not args:
                return tuple()
            ranges: tuple[tuple[float, float], ...] = tuple()
            skip: int = 0
            for i in range(len(args)):
                if skip > 0:
                    skip -= 1
                    continue
                current_range: tuple[float, float] = (float(args[i][0]), float(args[i][-1]))
                current_min: float = min(current_range)
                current_max: float = max(current_range)
                for r in args[1 + i:]:
                    if current_min <= min(r) <= current_max:
                        current_max = max(current_max, *r)
                        skip += 1
                ranges += ((current_min, current_max),)
            return ranges

        self.catalog.extend(sorted(catalog, key=lambda entry: entry[SPECIES_TAG]))
        squash_same_species_tag_entries()
        self.frequency_limits = merge_frequency_tuples(*self.frequency_limits, frequency_limits)


class Catalog:
    DEFAULT_SUFFIX: str = '.json.gz'

    class Opener:
        OPENERS_BY_SUFFIX: dict[str, Callable] = {
            '.json': open,
            '.json.gz': gzip.open,
            '.json.bz2': bz2.open,
            '.json.xz': lzma.open,
            '.json.lzma': lzma.open,
        }

        OPENERS_BY_SIGNATURE: dict[str, Callable] = {
            b'{': open,
            b'\x1F\x8B': gzip.open,
            b'BZh': bz2.open,
            b'\xFD\x37\x7A\x58\x5A\x00': lzma.open,
        }

        def __init__(self, path: str | PathLike[str]) -> None:
            self._path: Path = Path(path)
            suffix: str = self._path.suffix.casefold()
            self._opener: Callable
            if suffix in Catalog.Opener.OPENERS_BY_SUFFIX:
                self._opener = Catalog.Opener.OPENERS_BY_SUFFIX[suffix]
            else:
                if self._path.exists():
                    max_signature_length: int = max(map(len, Catalog.Opener.OPENERS_BY_SIGNATURE.keys()))
                    f: BinaryIO
                    with self._path.open('rb') as f:
                        init_bytes: bytes = f.read(max_signature_length)
                    key: bytes
                    value: Callable
                    for key, value in Catalog.Opener.OPENERS_BY_SIGNATURE.items():
                        if init_bytes.startswith(key):
                            self._opener = value
                            return

                raise ValueError(f'Unknown file: {path}')

        @contextmanager
        def open(self, mode: str, encoding: str | None = None,
                 errors: str | None = None, newline: str | None = None) -> TextIO | BinaryIO:
            """
            Open a file in a safe way. Create a temporary file when writing.

            See https://stackoverflow.com/a/29491523/8554611, https://stackoverflow.com/a/2333979/8554611
            """
            writing: bool = 'w' in mode.casefold()
            if encoding is None and 'b' not in mode.casefold():
                encoding = 'utf-8'
            tmp_path: Path = self._path.with_name(self._path.name + '.part')
            file: TextIO | BinaryIO
            with self._opener(tmp_path if writing else self._path, mode=mode,
                              encoding=encoding, errors=errors, newline=newline) as file:
                try:
                    yield file
                finally:
                    if writing:
                        file.flush()
                        os.fsync(file.fileno())
                        tmp_path.replace(self._path)

    def __init__(self, *catalog_file_names: str) -> None:
        self._data: CatalogData = CatalogData()
        self._sources: list[CatalogSourceInfo] = []

        for filename in catalog_file_names:
            if os.path.exists(filename) and os.path.isfile(filename):
                f_in: BinaryIO
                with Catalog.Opener(filename).open('rb') as f_in:
                    content: bytes = f_in.read()
                    try:
                        json_data: dict[str, list[float | None] | list[CatalogEntryType]] = json.loads(content)
                    except (json.decoder.JSONDecodeError, UnicodeDecodeError):
                        pass
                    else:
                        self._data.append(catalog=json_data[CATALOG],
                                          frequency_limits=(json_data[FREQUENCY][0],
                                                            (math.inf if json_data[FREQUENCY][1] is None
                                                             else json_data[FREQUENCY][1])))
                        build_datetime: datetime | None = None
                        if BUILD_TIME in json_data:
                            build_datetime = datetime.fromisoformat(cast(str, json_data[BUILD_TIME]))
                        self._sources.append(CatalogSourceInfo(filename=filename, build_datetime=build_datetime))

    def __bool__(self) -> bool:
        return bool(self._data.catalog)

    @property
    def is_empty(self) -> bool:
        return not bool(self)

    @property
    def sources(self) -> list[str]:
        source: CatalogSourceInfo
        return [source.filename for source in self._sources]

    @property
    def sources_info(self) -> list[CatalogSourceInfo]:
        return self._sources.copy()

    @property
    def catalog(self) -> list[CatalogEntryType]:
        return self._data.catalog

    @property
    def entries_count(self) -> int:
        return len(self._data.catalog)

    @property
    def frequency_limits(self) -> tuple[tuple[float, float], ...]:
        return self._data.frequency_limits if self._data.catalog else (0.0, math.inf)

    @property
    def min_frequency(self) -> float:
        return min(min(f) for f in self._data.frequency_limits) if self._data.frequency_limits else 0.0

    @property
    def max_frequency(self) -> float:
        return max(max(f) for f in self._data.frequency_limits) if self._data.frequency_limits else math.inf

    def filter(self, *,
               min_frequency: float = 0.0,
               max_frequency: float = math.inf,
               min_intensity: float = -math.inf,
               max_intensity: float = math.inf,
               temperature: float = -math.inf,
               any_name: str = '',
               any_formula: str = '',
               any_name_or_formula: str = '',
               anything: str = '',
               species_tag: int = 0,
               inchi_key: str = '',
               trivial_name: str = '',
               structural_formula: str = '',
               name: str = '',
               stoichiometric_formula: str = '',
               isotopolog: str = '',
               state: str = '',
               degrees_of_freedom: Optional[int] = None,
               timeout: Optional[float] = None
               ) -> list[CatalogEntryType]:
        """
        Extract only the entries that match all the specified conditions

        :param float min_frequency: the lower frequency [MHz] to take.
        :param float max_frequency: the upper frequency [MHz] to take.
        :param float min_intensity: the minimal intensity [log10(nm²×MHz)] to take.
        :param float max_intensity: the maximal intensity [log10(nm²×MHz)] to take, use to avoid meta-stable substances.
        :param float temperature: the temperature to calculate the line intensity at,
                                  use the catalog intensity if not set.
        :param str any_name: a string to match the ``trivialname`` or the ``name`` field.
        :param str any_formula: a string to match the ``structuralformula``, ``moleculesymbol``,
                                ``stoichiometricformula``, or ``isotopolog`` field.
        :param str any_name_or_formula: a string to match any field used by :param:any_name and :param:any_formula.
        :param str anything: a string to match any field at all.
        :param int species_tag: a number to match the ``speciestag`` field.
        :param str inchi_key: a string to match the ``inchikey`` field.
                              See https://iupac.org/who-we-are/divisions/division-details/inchi/ for more.
        :param str trivial_name: a string to match the ``trivialname`` field.
        :param str structural_formula: a string to match the ``structuralformula`` field.
        :param str name: a string to match the ``name`` field.
        :param str stoichiometric_formula: a string to match the ``stoichiometricformula`` field.
        :param str isotopolog: a string to match the ``isotopolog`` field.
        :param str state: a string to match the ``state`` or the ``state_html`` field.
        :param int degrees_of_freedom: 0 for atoms, 2 for linear molecules, and 3 for nonlinear molecules.
        :param float timeout: if positive, the maximum time [seconds] for filtering.
        :return: a list of substances with non-empty lists of absorption lines that match all the conditions.
        """

        if self.is_empty:
            return []

        def filter_by_frequency_and_intensity(catalog_entry: CatalogEntryType) -> CatalogEntryType:
            def intensity(_line: LineType) -> float:
                if catalog_entry[DEGREES_OF_FREEDOM] >= 0 and temperature > 0. and temperature != T0:
                    return (_line[INTENSITY] +
                            ((0.5 * catalog_entry[DEGREES_OF_FREEDOM] + 1.0) * math.log(T0 / temperature)
                             - ((1 / temperature - 1 / T0) * _line[LOWER_STATE_ENERGY] * 100. * h * c / k)) / M_LOG10E)
                else:
                    return _line[INTENSITY]

            new_catalog_entry: CatalogEntryType = catalog_entry.copy()
            if LINES in new_catalog_entry and new_catalog_entry[LINES]:
                min_frequency_index: int = search_sorted(min_frequency, new_catalog_entry[LINES],
                                                         key=lambda line: line[FREQUENCY]) + 1
                max_frequency_index: int = search_sorted(max_frequency, new_catalog_entry[LINES],
                                                         key=lambda line: line[FREQUENCY], maybe_equal=True)
                new_catalog_entry[LINES] = \
                    [line for line in new_catalog_entry[LINES][min_frequency_index:(max_frequency_index + 1)]
                     if min_intensity <= intensity(line) <= max_intensity]
            else:
                new_catalog_entry[LINES] = []
            return new_catalog_entry

        if (min_frequency > max_frequency
                or min_frequency > self.max_frequency
                or max_frequency < self.min_frequency):
            return []
        start_time: float = time.monotonic()
        if (species_tag or inchi_key or trivial_name or structural_formula or name or stoichiometric_formula
                or isotopolog or state or degrees_of_freedom or any_name or any_formula or any_name_or_formula
                or anything):
            trivial_name: str = trivial_name.casefold()
            name: str = name.casefold()
            any_name: str = any_name.casefold()
            any_name_or_formula_lowercase: str = any_name_or_formula.casefold()
            anything_lowercase: str = anything.casefold()
            selected_entries: list[CatalogEntryType] = []
            for entry in self.catalog:
                if timeout is not None and 0.0 < timeout <= time.monotonic() - start_time:
                    break
                if ((not species_tag or entry.get(SPECIES_TAG, 0) == species_tag)
                        and (not inchi_key or entry.get(INCHI_KEY, '') == inchi_key)
                        and (not trivial_name or entry.get(TRIVIAL_NAME, '').casefold() == trivial_name)
                        and (not structural_formula or entry.get(STRUCTURAL_FORMULA, '') == structural_formula)
                        and (not name or entry.get(NAME, '').casefold() == name)
                        and (not stoichiometric_formula
                             or entry.get(STOICHIOMETRIC_FORMULA, '') == stoichiometric_formula)
                        and (not isotopolog or entry.get(ISOTOPOLOG, '') == isotopolog)
                        and (not state or state in (entry.get(STATE, ''), entry.get(STATE_HTML, '')))
                        and (degrees_of_freedom is None or entry.get(DEGREES_OF_FREEDOM, -1) == degrees_of_freedom)
                        and (not any_name
                             or any_name in (entry.get(TRIVIAL_NAME, '').casefold(),
                                             entry.get(NAME, '').casefold(),))
                        and (not any_formula
                             or any_formula in (entry.get(STRUCTURAL_FORMULA, ''),
                                                entry.get(MOLECULE_SYMBOL, ''),
                                                entry.get(STOICHIOMETRIC_FORMULA, ''),
                                                entry.get(ISOTOPOLOG, ''),))
                        and (not any_name_or_formula
                             or any_name_or_formula_lowercase in (entry.get(TRIVIAL_NAME, '').casefold(),
                                                                  entry.get(NAME, '').casefold(),)
                             or any_name_or_formula in (entry.get(STRUCTURAL_FORMULA, ''),
                                                        entry.get(MOLECULE_SYMBOL, ''),
                                                        entry.get(STOICHIOMETRIC_FORMULA, ''),
                                                        entry.get(ISOTOPOLOG, ''),))
                        and (not anything
                             or anything in (str(entry[key]) for key in entry if key != LINES)
                             or anything_lowercase in (entry.get(TRIVIAL_NAME, '').casefold(),
                                                       entry.get(NAME, '').casefold(),))
                ):
                    filtered_entry: CatalogEntryType = filter_by_frequency_and_intensity(entry)
                    if filtered_entry[LINES]:
                        selected_entries.append(filtered_entry)
        else:
            filtered_entries: list[CatalogEntryType] = [
                filter_by_frequency_and_intensity(entry)
                for entry in self._data.catalog
                if timeout is None or (timeout > 0.0 and timeout >= time.monotonic() - start_time)
            ]
            selected_entries = [entry for entry in filtered_entries if entry[LINES]]
        return selected_entries

    def print(self, **kwargs: None | int | float | str) -> None:
        """
        Print a table of the filtered catalog entries

        :param kwargs: all arguments that are valid for :func:`filter <catalog.Catalog.filter>`
        :return: nothing
        """
        entries: list[CatalogEntryType] = self.filter(**kwargs)
        if not entries:
            print('nothing found')
            return

        names: list[str] = []
        frequencies: list[float] = []
        intensities: list[float] = []
        for entry in entries:
            for line in cast(LinesType, entry[LINES]):
                names.append(entry[NAME])
                frequencies.append(line[FREQUENCY])
                intensities.append(line[INTENSITY])

        def max_width(items: list[Any]) -> int:
            return max(len(str(item)) for item in items)

        names_width: int = max_width(names)
        frequencies_width: int = max_width(frequencies)
        intensities_width: int = max_width(intensities)
        for j, (n, f, i) in enumerate(zip(names, frequencies, intensities)):
            print(f'{n:<{names_width}} {f:>{frequencies_width}} {i:>{intensities_width}}')

    @staticmethod
    def save(filename: str,
             catalog: list[CatalogEntryType],
             frequency_limits: tuple[float, float] = (0.0, math.inf),
             build_time: datetime | None = None) -> None:
        if build_time is None:
            build_time = datetime.now(tz=timezone.utc)
        data_to_save: dict[str, list[CatalogEntryType] | tuple[float, float] | str] = {
            CATALOG: catalog,
            FREQUENCY: list(frequency_limits),
            BUILD_TIME: build_time.isoformat(),
        }
        opener: Catalog.Opener
        try:
            opener = Catalog.Opener(filename)
        except ValueError:
            opener = Catalog.Opener(filename + Catalog.DEFAULT_SUFFIX)

        def ensure_bytes(data: AnyStr) -> bytes:
            if isinstance(data, str):
                return data.encode('utf-8')
            if isinstance(data, bytes):
                return data
            raise TypeError('Unknown conversion to bytes')

        f: BinaryIO
        with opener.open('wb') as f:
            f.write(ensure_bytes(json.dumps(data_to_save, indent=4)))
